# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a high level API to SQL DB servers. It sits above
the Drizzle dbapi providing boots with a simple and easy to use interface to
drizzle.

This API can be implemented for any dbapi implementation allowing boots to
speak to a variety of DB servers."""

import time
import warnings
from collections import namedtuple
from boots.api import errors
from boots.api.nodes.node import Status
from boots.api import _, n_

_server_classes = []

# Named tuple to represent a columns description as provided by the dbapi.
ColumnDescription = namedtuple("ColumnDescription", 
                               ("name", "type_code", "display_size",
                                "internal_size", "precision", "scale",
                                "null_ok"))

class RowDescription:
    """Maintain descriptions for each column in a row."""
    def __init__(self, description):
        # Create a list of ColumnDescriptions for each column in the description.
        self.description = [ColumnDescription(*column) for column in description]
        # Create dict that maps column names to column index. eg "foo" -> 1.
        self.column_index = dict(zip((column[0] for column in description),
                                     range(len(description))))
    
    def index_by_name(self, name):
        """Return a columns description given its name."""
        return self.column_index[name]
     
    def __getitem__(self, key):
        """Return a column's description specified either by column name or
        by column index in the [] operator."""
        if type(key) is int:
            return self.description[key]
        elif type(key) is str:
            return self.description[self.column_index[key]]
        else:
            raise TypeError(_("Key must be a column index or column name"))
    
    def __iter__(self):
        return iter(self.description)
    
class Rows(list):
    """Represents a chunk of rows."""
    def __init__(self, sequence, description):
        super(Rows, self).__init__(sequence)
        self.description = description

class ResultInfo(Status):
    pass

class DurationCounter(object):
    """Class used to collect timing data. Execute code that needs to be timed
    in a with DurationCounter_Obj: block.""" 
    def __init__(self):
        self.duration = 0
        self._start_time = None
        
    def start(self):
        if self._start_time is None:
            self._start_time = time.time()
    
    def pause(self):
        if self._start_time is not None:
            self.duration += time.time() - self._start_time
            self._start_time = None
        
    def __enter__(self):
        self.start()
        return self
        
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pause()
        return False

class Server(object):
    """Represents a connection to a database server.
    
    This object is used to execute commands on a remote server."""

    def __init__(self, dbapi_module, hostname, port, database, auth):
        """Initialize a Server object.
        
        hostname should be a string.
        port should be a positive integer value.
        auth should be a dictionary possibly containing keys:
        'username', 'password', and 'database' mapping to their
        appropriate string values."""
        self.db = dbapi_module    
        self.hostname = hostname
        self.port = port
        self.database = database
        self.auth = auth
        self._connection = None

    def __str__(self):
        """Convert a Server object to a str. Results in 'hostname:port'."""
        return "{0}:{1}".format(self.hostname, self.port)

    def _connect(self):
        return self.db.connect(
            host     =  self.hostname,  
            port     =  self.port,
            database =  self.database,
            username =  self.auth.get("username") or "",
            password =  self.auth.get("password") or "")

    def connect(self):
        """Establishes database connection.

        Connect to the database server whose details were provided through
        the __init__ method."""
        try:
            self._connection = self._connect()
        except self.db.Error as e:
            raise errors.ConnectionError(_("Could not connect to {0}").format(self),
                                         self._wrap_exception(e))

    @property
    def is_connected(self): 
        """Describes connection state.

        TRUE if connected, FALSE if disconnected."""
        return self._connection is not None

    def _check_connected(self):
        """Raise a ProgrammingError if the Connection has been closed."""
        if not self.is_connected:
            raise errors.ConnectionError(_("Connection closed"))

    def disconnect(self):
        """Disconnect from the connected database server."""
        self._check_connected()
        self._connection.close()
        self._connection = None
    
    def _wrap_exception(self, e):
        if isinstance(e, self.db.DatabaseError):
            return errors.DatabaseError(e.value, e)
        else:
            return e
    
    def _execute(self, query):
        cursor = self._connection.cursor()
        try:
            cursor.execute(query)
        except self.db.Error as e:
            raise self._wrap_exception(e)
        return cursor
    
    def execute(self, query):
        """Executes a query, returning an iterator of resulting rows."""
        self._check_connected()
        begin_time = time.time()
        server_timer = DurationCounter()
        
        # Catch all warnings so that they may be passed down the node graphs.
        with warnings.catch_warnings(record=True) as warns:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            try:
                with server_timer:
                    cursor = self._execute(query)
            # If we have lost our connection (timed out, etc), attempt to
            # reconnect to the server and continue.
            except errors.LostConnectionError as lost_e:
                # FIXME: this is NOT thread safe
                # Attempt to reconnect and retry
                yield Status(_("Lost connection to {0}. Reconnecting...").format(self))
                
                if self.is_connected:
                    self.disconnect()
                                
                try:
                    self.connect()
                except Exception as e:
                    lost_e.reconnect_exception = e
                    raise lost_e
                else:
                    yield Status(_("Reconnected."))
    
                with server_timer:
                    cursor = self._execute(query)
            
            if cursor.description:
                while True:
                    with server_timer:
                        chunk = cursor.fetchmany(256)
                    # FIXME: Eventually it would be good to implement chunk packeting
                    # as a generic nodes feature.
                    if chunk:
                        yield Rows(chunk, RowDescription(cursor.description))
                    else:
                        break
        
        # Pass the caught warnings down the node graph.
        for warning in warns:
            yield warning.message
                
        # Pass a final information packet populated with information possibly
        # unknown until all rows have passed through the dbapi.
        yield ResultInfo({"description": RowDescription(cursor.description) 
                                         if cursor.description else None,
                          "row_count": cursor.rowcount,
                          "server_elapsed": server_timer.duration,
                          "begin_time": begin_time,
                          "end_time": time.time()})
        return
                
    @property
    def server_version(self):
        self._check_connected()
        return self._connection.server_version
    
    @property
    def protocol_version(self):
        self._check_connected()
        return self._connection.protocol_version

# If the drizzle dbapi module exists import it then derive the Server object
# modifying it to accomodate the drizzle dbapi.
try:
    import drizzle.db, drizzle.errors
    class LibDrizzleServer(Server):
        def __init__(self, hostname, port, database, auth):
            Server.__init__(self, drizzle.db, hostname, port, database, auth)
        
        def _connect(self):
            connection = Server._connect(self)
            connection.autocommit = True
            return connection
        
        def _wrap_exception(self, e):
            if isinstance(e, drizzle.errors.DatabaseError):
                return errors.CodedDatabaseError(e.code, e.msg, e)
            else:
                return e
        
        def _execute(self, query):
            try:
                cursor = self._connection.cursor(convert=False)
                cursor.execute(query)
            except drizzle.errors.LostConnectionError:
                raise errors.LostConnectionError(server=self)
            except drizzle.errors.DatabaseError as e:
                raise self._wrap_exception(e)
            else:
                return cursor
            
            
except ImportError:
    LibDrizzleServer = None
else:
    _server_classes.append(LibDrizzleServer)

# If the MySQLdb dbapi module exists import it then derive the Server object
# modifying it to accomodate the MySQLdb dbapi.
try:
    import MySQLdb
    import MySQLdb.converters
    class MySQLdbServer(Server):
        def __init__(self, hostname, port, database, auth):
            Server.__init__(self, MySQLdb, hostname, port, database, auth)

        def _connect(self):
            """Establishes database connection.

            Connect to the database server whose details were provided through
            the __init__ method."""

            converter = self.db.converters.conversions.copy()
            for x in range(0,256):
                converter.pop(x, None)

            connection = self.db.connect(
                host     =  self.hostname,  
                port     =  self.port,
                db       =  self.database if self.database is not None else "",
                user     =  self.auth.get("username") or "",
                passwd   =  self.auth.get("password") or "",
                conv     =  converter)
            
            connection.autocommit(True)
            return connection

        def _wrap_exception(self, e):
            if isinstance(e, MySQLdb.DatabaseError):
                if len(e.args) == 2:
                    return errors.CodedDatabaseError(e.args[0], e.args[1], e)
                else:
                    return errors.DatabaseError(e.args[0], e)
            else:
                return e

        @property
        def server_version(self):
            self._check_connected()
            return self._connection.get_server_info()
        
except ImportError:
    MySQLdbServer = None
else:
    _server_classes.append(MySQLdbServer)

# Create a Server object capable of speaking to a Drizzle server. We can use
# libdrizzle (preferred) or MySQLsb.
def DrizzleServer(hostname, port, database, auth):
    if LibDrizzleServer is not None:
        return LibDrizzleServer(hostname, port, database, auth)
    elif MySQLdbServer is not None:
        return MySQLdbServer(hostname, port, database, auth)
    else:
        raise errors.BootsError(_("MySQLdb or python-libdrizzle are required to connect to Drizzle databases."))

class Query(str):
    def substitute(self, rows): pass
    
