# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import sys
import exceptions
import traceback

from boots.api import _, n_

def _format_exception_line(e):
    if e:
        lines = traceback.format_exception_only(type(e), e)
        text = ". ".join(line.strip() for line in lines)
        return text

class BootsError(exceptions.StandardError):
    """Base class of boots error exceptions."""
    pass

class BootsWarning(exceptions.Warning):
    """Base class of boots warnings."""
    pass

class CausedException(BootsError):
    """Errors caused by other errors, optionally storing a traceback."""
    # Note: This will eventually be superseded by new features in the exception
    # system of Python 3: __context__, __cause__, and __traceback__
    def __init__(self, cause=None, traceback=None):
        BootsError.__init__(self)
        self.cause = cause
        self.traceback = traceback
        
    @classmethod
    def from_exception(cls, e):
        last_e, last_traceback = sys.exc_info()[1:]
        return cls(e, last_traceback if e is last_e else None)
    
    def __str__(self):
        return str(self.cause)
    
    def format_exception(self):
        if self.traceback:
            return "".join(traceback.format_exception(type(self.cause),
                                                      self.cause,
                                                      self.traceback))
        
    def format_exception_line(self):
        if self.cause:
            return _format_exception_line(self.cause)

class ServerError(BootsError):
    """Base class for Server exceptions."""
    pass

class DatabaseError(ServerError, CausedException):
    """Errors returned from database servers."""
    def __init__(self, msg, cause):
        ServerError.__init__(self)
        CausedException.__init__(self, cause)
        self.msg = msg
        
class CodedDatabaseError(DatabaseError):
    def __init__(self, code, msg, cause):
        DatabaseError.__init__(self, msg, cause)
        self.code = code
        
    def __str__(self):
        return "{msg} (#{code})".format(msg=self.msg, code=self.code)
    
class ConnectionError(ServerError, CausedException):
    """Errors related to server connections."""
    def __init__(self, msg, cause=None):
        CausedException.__init__(self, cause)
        self.msg = msg
        
    def __str__(self):
        if self.cause is not None:
            return "{msg} -- {exc}".format(msg=self.msg, exc=self.cause)
        else:
            return self.msg

class LostConnectionError(ConnectionError):
    """Raised when a connection is lost."""
    def __init__(self, msg=None, server=None, reconnect_exception=None):
        ConnectionError.__init__(self, msg)
        self.server = server
        self.reconnect_exception = reconnect_exception
    
    def __str__(self):
        msg = []
        if self.msg:
            msg.append(self.msg)
        else:
            msg.append(_("Lost connection to {0}").format(self.server))

        if self.reconnect_exception:
            msg.append(_("(Exception upon reconnect: {0})")
                       .format(_format_exception_line(self.reconnect_exception)))
            
        return " ".join(msg)
    