# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a class that builds a boots client configuration. The
configuration is built using a set of defaults, an rc file, and command line
options/arguments."""

import os, os.path
import optparse
from getpass import getpass

import info
from boots.app import _, n_

def find_executable(name):
    """Finds an executable in the search path and returns the full path."""
    # Search algorithm adapted from os._execvpe 
    head, tail = os.path.split(name)
    if head and os.path.exists(name):
        return name
    else:
        if 'PATH' in os.environ:
            envpath = os.environ['PATH']
        else:
            envpath = os.defpath
        PATH = envpath.split(os.pathsep)
        for dir in PATH:
            fullname = os.path.join(dir, name)
            if os.path.exists(fullname):
                return fullname

def get_auto_pager():
    """Returns a preferred pager command.
    
    If less is available it is used with options to not page if input fits on
    the screen (-F) and to leave output on the screen after exit (-X). If less
    is not available more is used."""
    if find_executable("less"):
        return "less -XF"
    elif find_executable("more"):
        return "more"    

def optparse_password(option, opt_str, value, parser):
    assert value is None
    
    password = None
    if len(parser.rargs) > 0:
        arg = parser.rargs[0]
        # FIXME: the optparse documentation says we should ignore these two
        # conditions, but is that appropriate behavior? Will this surprise
        # a user whose password starts with "-"?
        # FIXME: This should be moved into the driver eventually.
        if not arg.startswith("--") and not arg.startswith("-"):
            password = parser.rargs.pop(0)
            
    if not password:
        try:
            password = getpass()
        except (KeyboardInterrupt, EOFError):
            password = None
        
    setattr(parser.values, option.dest, password)


class ClientConfig(object):
    """Class used to read client configuration from files and the cli."""

    def __init__(self, conf_file_loc = None):
        """Initialize a configuration. Optionally takes a path to the rc file.

        This method performs no IO and after initializaion you must call load
        to populate a config."""
        # Specify default values for a client configuration.
        self._defaults = {"command":            None,
                          "database":           None,
                          "script":             None,
                          "lingo":              "sql",
                          "debug":              False,
                          "quiet":              False,
                          "rcfile":             info.RCFILE,
                          "host":               "localhost",
                          "port":               3306,
                          "username":           os.getenv("USER"),
                          "password":           None,
                          "prompt1":            "> ",
                          "prompt2":            "+ ",
                          "pager":              None,
                          "pager_command":      None,
                          "terminating_char":   ";",
                          "vertical_output":    False,
                          "history_length":     100,
                          "history_file":       os.path.expanduser("~/.boots_history")}
        self._dict = self._defaults.copy()

        self.conf_file = conf_file_loc

        # Build cli option parser.
        version_string = "{name} version {version}".format(name = info.NAME,
                                                           version = info.VERSION)
        self._cli_parser = optparse.OptionParser(version = version_string,
                                                 conflict_handler="resolve");

        self._cli_parser.add_option('--debug',
                                    action = 'store_true',
                                    dest = 'debug',
                                    help = _('turn on debugging output'))
        self._cli_parser.add_option('-q', '--quiet',
                                    action = 'store_true',
                                    dest = 'quiet',
                                    help = _('suppress warning output'))
        self._cli_parser.add_option("-r", "--rc",
                                    action = "store",
                                    type = "string",
                                    dest = "rcfile",
                                    help = _("filename of user configuration file"))
        self._cli_parser.add_option("-X", "--norc",
                                    action = "store_true",
                                    dest = "norc",
                                    default = False,
                                    help = _("do not read user configuration file"))

        command_group = optparse.OptionGroup(self._cli_parser, "Command Processing Options")
        command_group.add_option("-c", "--command",
                                 action = "store",
                                 type = "string",
                                 dest = "command",
                                 help = _("execute command and exit"))
        command_group.add_option("-f", "--script",
                                 action = "store",
                                 type = "string",
                                 dest = "script",
                                 help = _("execute commands from file and exit"))
        command_group.add_option('-l', '--lingo',
                                 action = 'store',
                                 type = 'string',
                                 dest = 'lingo',
                                 help = _('set the lingo to use'))
        self._cli_parser.add_option_group(command_group)
       
        server_group = optparse.OptionGroup(self._cli_parser, "Server Options")
        server_group.add_option("-h", "--host",
                                action = "store",
                                type = "string",
                                dest = "host",
                                help = _("connect to host"))
        server_group.add_option("-p", "--port",
                                action = "store",
                                type = "int",
                                dest = "port",
                                help = _("use port number"))
        server_group.add_option("-u", "--user",
                                action = "store",
                                type = "string",
                                dest = "username",
                                help = _("login with username"))
        server_group.add_option("-P", "--password",
                                action = "callback",
                                callback = optparse_password,
                                dest = "password",
                                help = _("connect using password. If none is given, query for password"))
        server_group.add_option("-D", "--database",
                                 action = "store",
                                type = "string",
                                dest = "database",
                                help = _("use specified database"))
        self._cli_parser.add_option_group(server_group)

        misc_group = optparse.OptionGroup(self._cli_parser, "Miscellaneous Options")
        misc_group.add_option("--pager",
                              action = "store",
                              type = "string",
                              dest = "pager",
                              help = _("pipe query results to the specified pager."))
        misc_group.add_option("-t", "--terminatingchar",
                              action = "store",
                              type = "string",
                              dest = "terminating_char",
                              help = _("specify the SQL statement terminating character (default is ';')"))
        misc_group.add_option("--vertical-output",
                              action = "store_true",
                              dest = "vertical_output",
                              default = False,
                              help = _("print vertical output (like \G)"))
        misc_group.add_option("-F", "--historyfile",
                              action = "store",
                              type = "string",
                              dest = "history_file",
                              help = _("specify file to save history to"))
        misc_group.add_option("-L", "--historylength",
                              action = "store",
                              type = "int",
                              dest = "history_length",
                              help = _("specify max history file length"))
        self._cli_parser.add_option_group(misc_group)
 
    def __getitem__(self, key):
        """Allows one to use the [] operator to get config items."""
        return self._dict[key]

    def __setitem__(self, key, value):
        """Allows one to use the [] to set config items."""
        self._dict[key] = value

    @property
    def options(self):
        """Give entire configuration as a dictionary."""
        return self._dict

    def load(self):
        """Load a configuration from the cli and rc file.

        Options specified in rc file overwrite defaults and cli options
        overwrite rc options."""
        from_cli = self.get_cli_conf()
        
        # Only load config options from the file if the norc option was not set
        # on the cli. Additionally if the conf_file is not already set use the
        # file specified on the cli if available and the default file
        # otherwise.
        if not from_cli["norc"]:
            if not self.conf_file:
                if "rcfile" in from_cli:
                    self.conf_file = from_cli["rcfile"]
                else:
                    self.conf_file = self._defaults["rcfile"]
            from_file = self.get_file_conf(self.conf_file)
            self._dict.update(from_file)
            
        self._dict.update(from_cli)
        self._interpret_options()
 
    def _interpret_options(self):
        # Interpret options
        if self["pager"] == "auto":
            self["pager_command"] = get_auto_pager()
        elif self["pager"]:
            self["pager_command"] = self["pager"]

    def get_file_conf(self, filepath):
        """Read a configuration from the specified file. Return a dict."""
        file_dict = {}
        try:
            # TODO is there a better/safer way to easily parse a config file?
            execfile(filepath, {}, file_dict)
        except IOError:
            if filepath != info.RCFILE:
                # Only print an error if a non-default rc file was specified.
                print(_("rcfile {0} not found").format(filepath))
        except SyntaxError:
            print(_("rcfile {0} contains a syntax error").format(filepath))
        except Exception as e:
            print(_("rcfile {0} contains an error: {1}").format(filepath, e))
        return file_dict

    def get_cli_conf(self):
        """Read a configuration from the cli. Return a dict."""
        (options, args) = self._cli_parser.parse_args()
        tmp_options = vars(options)
        new_options = {}

        # We want to remove keys associated with None. This allows us to safely
        # merge the default config values with those specified on the cli.
        for key, value in tmp_options.items():
            if value != None:
                new_options[key] = value
        self._args = args
        return new_options

    def update(self, pairs_to_add):
        """Update configuration with provided dictionary."""
        self._dict.update(pairs_to_add)
